=begin
###
Copyright 2011, TIG (c)
Permission to use, copy, modify, and distribute this software for any 
purpose, and currently without fee, is hereby granted, provided that  
this text and the above copyright (c) notice appear in all copies.
This software is provided "as is" and without any express or implied 
warranties, including, without limitation, the implied warranties of 
merchantability and fitness for a particular purpose.
###
Name: TIG-Superdrape.rb
###
Activate:
Menu > 'Tools' > 'Super Drape'
OR
Menu > View > Toolbar > 'Super Drape'
OR
Ruby-Console/Code: TIG-superdrape
###
###
Usage:
This tool drapes one group onto another group - much like the 
Sandbox Drap tool [BUT much more consistently] AND it also copies materials 
from with the 1st group to the 2nd group.
Activate the tool.
You are prompted to select the group to drape, and then the 2nd group to 
drape onto.
It is one step undoable.
###
Tips:
The draped group must be above the 2nd group, but placing it very far above 
it might result in some inaccuracies.
The two groups should consist of one set of faces, oriented 'up'.
The lowest face is the draped group will supply the materials if there are 
'layers' of faces that overlap.
The highest face will receive the drape if there are 'layers' of faces that 
overlap; to be on the safe side it's recommended you use single 'layers' 
of non-overlapping faces in the groups.
Any face < ~1mm will probably be ignored when materials are assigned.
Faces may be angled from the horizontal in both groups - inside a group 
or even a group itself.
Any textured materials in the draped group will have their UVs correctly 
transposed in the materials used in the 2nd group, but they might be 
foreshortened unexpectedly if a draped group face slopes significantly 
more than the equivalent face in the 2nd group.
Any processed groups with more than one instance will be made unique first.
###
Donations: PayPal.com to info @ revitrev.org
###
Version:
1.0 20111004 First issue.
1.1 20111005 Face 'pairing' more robust. Only uppermost face gets draped.
1.2 20111006 Trapped 'group.material'. Robust Intersecting/overlap-trimming.
1.3 20111006 Group.copy now recast to avoid rogue scripts messing it up.
             Multi-level faces in the draped group are now allowed.
1.4 20111008 1st group temporarily moved nearer to 2nd to minimize errors.
             Original is unaffected by group.material >> default faces.
1.5 20111010 Adjusted to minimize missed facets etc.
###
=end
###
require 'sketchup.rb'
###
module TIG
 class TIG::Superdrape
  def initialize()
    @ip  = nil
    @ip1 = nil
  end
  def activate()
    @model=Sketchup.active_model
    @ss=@model.selection
    @ss.clear
    @ip  = Sketchup::InputPoint.new
    @ip1 = Sketchup::InputPoint.new
    @count=0
    @draper=nil
    @drapee=nil
    @msg="Super Drape: Select a Group to Drape..."
    self.reset()
  end
  def deactivate(view)
    view.invalidate
  end
  def onCancel(flag, view)
    view.invalidate
    Sketchup.send_action("selectSelectionTool:")
    return nil
  end
  def resume(view)
    Sketchup::set_status_text(@msg, SB_PROMPT)
  end
  def reset()
    Sketchup::set_status_text(@msg, SB_PROMPT)
  end
  def onMouseMove(flags, x, y, view)
    @ip.pick(view, x, y)
    if @ip != @ip1
        view.invalidate if @ip.display? or @ip1.display?
        @ip1.copy!(@ip)
        view.tooltip = @ip1.tooltip
    end
  end
  def onLButtonDown(flags, x, y, view)
    if @ip1.valid?
        @pt=@ip1.position
        ph = view.pick_helper
        ph.do_pick(x,y)
        all = ph.all_picked
        return nil if not all
        @sel=nil
        all.each{|e|
          if e.class==Sketchup::Group
            @sel=e
            break
          end
        }
        self.set() if @sel
        self.drape() if @count>1
    end
  end
  def set()
    (UI.messagebox("Super Drape: Pick a DIFFERENT Group!");return nil) if @sel==@draper
    (UI.messagebox("Super Drape: Groups MUST be in 'active_entities'!");return nil) if @sel.parent.entities!=@model.active_entities
    if @count==0
      @draper=@sel
      @ss.add(@draper)
      @count=1
      @msg="Super Drape: Select Another Group to Drape onto..."
      Sketchup::set_status_text(@msg, SB_PROMPT)
    else
      @drapee=@sel
      @ss.add(@drapee)
      @model.active_view.invalidate
      @count=2
      @msg="Super Drape: Calculating."
      Sketchup::set_status_text(@msg, SB_PROMPT)
    end
  end
  def drape()
    ###check draper is above drapee
    minZ=@draper.bounds.min.z
    maxZ=@drapee.bounds.max.z + 1.mm
    if maxZ>minZ
      UI.messagebox("Super Drape: The 1st Selected Group MUST be ABOVE the 2nd!\nTry again...")
      self.activate
      return nil
    end
    begin
      @model.start_operation("Super Drape", true)
    rescue
      @model.start_operation("Super Drape")
    end
    ###   
    @drapee.make_unique if @drapee.entities.parent.instances[1]
    ###
    min=@draper.bounds.min
    max=min.clone
    max.z=@drapee.bounds.max.z + 10.mm
    vec=min.vector_to(max)
    defn=@draper.entities.parent
    draper=@model.active_entities.add_instance(defn, @draper.transformation)
    draper.make_unique if draper.entities.parent.instances[1]
    tt=Geom::Transformation.translation(vec)
    draper.transform!(tt)
    td=draper.transformation
    group=@model.active_entities.add_instance(defn, td)
    group.make_unique if group.entities.parent.instances[1]
    tg=group.transformation
    ###
    if @draper.material
      defs=false
      @draper.entities.each{|e|
        if e.class==Sketchup::Face and not e.material
          defs=true
          break
        end
      }
      if defs
        mna=@draper.material.display_name
        if UI.messagebox("Super Draper: The Drape Group has a Material\n\""+mna+"\"\nDo you want to use that for any <Default> Faces ?",MB_YESNO,"")==6 ### 6=YES 7=NO
          draper.entities.each{|e|e.material=@draper.material if e.class==Sketchup::Face and not e.material}
        end
      end
    end
    ###
    orig=@draper.transformation.origin
    norig=orig.clone
    norig.z=norig.z+100.m
    vtp=orig.vector_to(norig)
    ttp=Geom::Transformation.translation(vtp)
    @draper.transform!(ttp)
    @draper.hidden=true
    begin
      @model.active_view.refresh
    rescue 
      @model.active_view.invalidate
    end
    ###
    @ss.clear
    ###
p1=draper.bounds.max
p2=p1.clone
p2.z=@drapee.bounds.min.z-10.mm
ve=p1.vector_to(p2)
ve.transform!(tg.inverse)
###
edges=[]
gents=group.entities
gents.each{|e|
  next unless e.class==Sketchup::Edge
  e.hidden=false
  e.smooth=false
  e.soft=false
  edges << e
}
###
@msg=@msg+' .';Sketchup::set_status_text(@msg, SB_PROMPT)
###
###
tpts=[]
bedges=[]
edges.each{|e|
  tpts << e.start.position.to_a
  tpts << e.end.position.to_a
  ps=e.start.position.offset(ve)
  pe=e.end.position.offset(ve)
  bedges << gents.add_line(ps, pe)
}
tpts.uniq!
edges.each_with_index{|e,i|
  gents.add_line(e.start.position, bedges[i].start.position)
  gents.add_line(e.end.position, bedges[i].end.position)
}
edges.each{|e|e.find_faces}
@msg=@msg+' .';Sketchup::set_status_text(@msg, SB_PROMPT)
###
#=begin
### intersect with drapee
nedges=gents.intersect_with(true, group.transformation, gents, group.transformation, true, @drapee)
@msg=@msg+' .';Sketchup::set_status_text(@msg, SB_PROMPT)
### intersect with self 
gents.intersect_with(true, group.transformation, gents, group.transformation, true, gents.to_a)
@msg=@msg+' .';Sketchup::set_status_text(@msg, SB_PROMPT)
fedges=[]
gents.each{|e|fedges << e if e.class==Sketchup::Edge and not e.faces[0]}
### remove unneeded vertical lines
togos=[]
gents.each{|e|
  next unless e.class==Sketchup::Edge
  next unless e.line[1].parallel?(ve)
  if tpts.include?(e.start.position.to_a)
    next unless e.length==e.start.position.distance(e.start.position.offset(ve))
  end
  togos << e
}
gents.erase_entities(togos)
### remake missing faces in spanned edges
gents.to_a.each{|e|
  next unless e.class==Sketchup::Edge and not e.faces[0]
  conn=false
  (e.start.edges+e.end.edges).each{|ee|
    if ee.line[1].parallel?(ve)
      conn=true
      break
    end
  }
  next if not conn
  if not e.start.faces[0]
    p0=e.start.position
    rt=@model.raytest([p0, Z_AXIS])
    next unless rt
    p1=rt[0]
    li=gents.add_line(p1, p0)
    li.find_faces
  elsif not e.end.faces[0]
    p0=e.end.position
    rt=@model.raytest([p0, Z_AXIS])
    next unless rt
    p1=rt[0]
    li=gents.add_line(p1, p0)
    li.find_faces
  end
}
### move down so only topmost intersects
vvv=ve.clone
vvv.length=5.mm
tz=Geom::Transformation.translation(vvv)
group.transform!(tz)
###
@msg=@msg+' .';Sketchup::set_status_text(@msg, SB_PROMPT)
###
#=end
t0=Geom::Transformation.new()
dt=@drapee.transformation
dents=@drapee.entities
tgp=dents.add_group()
tents=tgp.entities
###
    ###
    @msg="Super Drape: Making Outline..."
    Sketchup::set_status_text(@msg, SB_PROMPT)
    ###
    dents.intersect_with(true, dt, tents, dt, true, group)
    group.erase! if group.valid?
    tents.to_a.each{|e|e.erase! if e.line[1].parallel?(ve)}
    4.times{ dents.intersect_with(true, t0, dents, t0, true, dents.to_a) }
    tgp.erase! if tgp.valid?
    ###
    begin
      @model.active_view.refresh
    rescue 
      @model.active_view.invalidate
    end
    ###
#######
    dfaces=[]
    dents.each{|e|dfaces << e if e.class==Sketchup::Face}
    @msg="Super Drape: Adding Materials."
    ctr=0
    dfaces.each{|face|
      if ctr%50==0
        @msg=@msg+" ."
        begin
          @model.active_view.refresh
        rescue 
          @model.active_view.invalidate
        end
      end
      ctr+=1
      Sketchup::set_status_text(@msg, SB_PROMPT)
      ###
      p0=face.outer_loop.vertices[0].position
      p1=face.outer_loop.vertices[1].position
      p2=face.outer_loop.vertices[2].position
      v10=p1.vector_to(p0)
      v12=p1.vector_to(p2)
      up=v12.cross(v10)
      #nm=face.normal
      pt=p1.offset(v12, 1.mm)
      an=v12.angle_between(v10)/2.0
      #an = -an if nm!=up
      ptn=nil
      360.times{|i|
        begin
          tr=Geom::Transformation.rotation(p1, up, an+i.degrees)
          ptn=pt.transform(tr)
          ptn=ptn.project_to_plane(face.plane)
          if face.classify_point(ptn)==Sketchup::Face::PointInside
            break
          else
            ptn=nil
          end
        rescue
        end
      }
      next if not ptn
      ###
      ptn.transform!(dt)
      ###
      ray=@model.raytest([ptn, Z_AXIS])
      next if not ray
      next if ray[1][0]==@drapee
      next if ray[1][0]!=draper
      fac=ray[1][-1]
      fac=fac.faces[0] if fac.class==Sketchup::Edge
      next if not fac or fac.class!=Sketchup::Face
      mat=fac.material
      tw=Sketchup.create_texture_writer
      ###
     if mat and mat.texture ### adjust UVs
      samples=[]
      samples << fac.vertices[0].position            #0,0 | Origin
      samples << samples[0].offset(fac.normal.axes.x)#1,0 | Offset Origin in X
      samples << samples[0].offset(fac.normal.axes.y)#0,1 | Offset Origin in Y
      samples << samples[1].offset(fac.normal.axes.y)#1,1 | Offset X in Y
      uvhelper=fac.get_UVHelper(true, true, tw)### get uvs
      xyz=[]
      uvf=[]
      samples.each{|pos|
        xyz << pos
        uvq = uvhelper.get_front_UVQ(pos)
        uvf << self.flattenUVQ(uvq)# Front
      }
      pts=[]
      (0..3).each{|i|
        pt=xyz[i].transform(td) 
        line=[pt, @drapee.transformation.zaxis.reverse]
        pt=Geom.intersect_line_plane(line, face.plane)
        pts << pt
        pts << uvf[i]
      }
      begin
        face.position_material(mat, pts, true)
      rescue
        face.material=mat
      end
     else ### it's a plain material
      face.material=mat
     end
      ###
    }
    draper.erase! if draper.valid?
    @draper.transform!(ttp.inverse)
    @draper.hidden=false
    ###
    UI.beep
    @model.commit_operation
    self.onCancel(0, @model.active_view)
  end
  def flattenUVQ(uvq)
    return Geom::Point3d.new(uvq.x / uvq.z, uvq.y / uvq.z, 1.0)
  end
 end#class
unless file_loaded?(__FILE__)
  cmd=UI::Command.new('Super Drape'){Sketchup.active_model.select_tool(TIG::Superdrape.new())}
  cmd.tooltip='Super Drape'
  cmd.status_bar_text='Super Drape: drapes 1st selected group over 2nd group, keeping materials.'
  UI.menu('Tools').add_item(cmd)
  cmd.small_icon='TIGtools/SuperDrape16x16.png'
  cmd.large_icon='TIGtools/SuperDrape24x24.png'
  toolbar=UI::Toolbar.new('Super Drape')
  toolbar.restore if toolbar.get_last_state==TB_VISIBLE
  toolbar.add_item(cmd)
end#unless
file_loaded(__FILE__) 
end
